import os

from fastapi import FastAPI, Depends, Request

from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from fastapi.responses import HTMLResponse
from metaworkspace.runtime.api import api
from metaworkspace.runtime.api import auth
from metaworkspace.runtime.workspace import mws
from starlette.middleware.cors import CORSMiddleware


DIR = os.path.realpath(os.path.dirname(__file__))
STATIC = os.path.join(DIR, "../../client", "static")
ASSETS = os.path.join(DIR, "../../client", "assets")
PROJECTS = mws.projects_dir
TEMPLATES = os.path.join(DIR, "../../client")


def create_app():
    """Create app wrapper to overcome middleware issues."""
    templates = Jinja2Templates(directory=TEMPLATES)

    app = FastAPI()
    #app.include_router(project.router)
    #app.include_router(auth.router)
    app.mount("/static", StaticFiles(directory=STATIC), name="static")
    app.mount("/assets", StaticFiles(directory=ASSETS), name="assets")
    app.mount("/api/data", StaticFiles(directory=PROJECTS), name="data")

    origins = [
        "http://localhost:3000",
        "http://127.0.0.1:5000",
        "http://localhost:5000"
    ]

    return CORSMiddleware(
        app=app,
        allow_origins=origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    ), app, templates


app, fastapi, templates = create_app()


fastapi.include_router(api.router)
fastapi.include_router(auth.router)
@fastapi.get("/{path:path}", response_class=HTMLResponse)
async def root(request: Request):
    return templates.TemplateResponse("index.html", {"request": request})






import logging
import os
from logging.handlers import RotatingFileHandler
from metaworkspace.runtime.workspace import mws 


def setup_logging(name, scrnlog=False, txtlog=True, loglevel=logging.DEBUG):
    print(f"Initializing logger {name}")
    logdir = os.path.abspath(mws.logs_dir)

    if not os.path.exists(logdir):
        os.mkdir(logdir)

    log = logging.getLogger(name)
    log.setLevel(loglevel)

    log_formatter = logging.Formatter("%(asctime)s - %(levelname)s :: %(message)s")

    if txtlog:
        txt_handler = RotatingFileHandler(os.path.join(logdir, f"{name}.log"))
        txt_handler.doRollover()
        txt_handler.setFormatter(log_formatter)
        log.addHandler(txt_handler)
        log.info(f"Logger {name} initialised.")

    if scrnlog:
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(log_formatter)
        log.addHandler(console_handler)




import os
import sys
import time
from multiprocessing import Process, Queue
from typing import List
from metaworkspace.runtime.logging import setup_logging
import logging
import fnmatch

from metaworkspace.runtime.processing.loader import load_job
from watchdog.events import PatternMatchingEventHandler
from watchdog.observers import Observer


def JobWorker(queue: Queue, id: str):
    logid = f'worker{id}'
    setup_logging(logid)
    log = logging.getLogger(logid)

    while True:
        job_dir = queue.get(block=True, timeout=None)
        log.info(f"Worker {id} started: {job_dir}")
        job = load_job(job_dir)
        try:
            job.run() 
            job.cleanup()
            log.info(f"Worker {id} done: {job_dir}")
        except Exception as e:
            log.error(f"Worker {id}: {job_dir}")
            log.error(e)


class JobQueue:
    def __init__(self):
        self.queue = Queue()
        self.workers: List[Process] = []

    def run_workers(self, worker_count):
        for i in range(worker_count):
            p = Process(target=JobWorker, args=(self.queue, i))
            p.start()
            self.workers.append(p)

    def stop_workers(self):
        for p in self.workers:
            p.terminate()

    def add_job(self, job):
        self.queue.put(job)

    @property
    def status(self):
        status = "Runners: "
        for i, w in enumerate(self.workers):
            status += f"[ worker {i} with pid {w.pid} is alive: {w.is_alive()} ], "
        status += f"queue [ full: {self.queue.full()} ] [empty: {self.queue.empty()} ]"
        return status


class EventManager:
    def __init__(self, queue: JobQueue):
        self.queue = queue
        setup_logging('events')

    def on_created(self, event):
        log = logging.getLogger('events')
        log.info(f"{event.src_path}")
        job_dir = os.path.dirname(event.src_path)
        self.queue.add_job(job_dir)


class JobManager:
    def __init__(self, job_directory):
        setup_logging('runners')
        self.patterns = ['job.ready']
        self.ignore_patterns = None
        self.ignore_directories = False
        self.case_sensitive = True
        self.go_recursively = True
        self.job_directory = job_directory
        self.queue = JobQueue()
        self.em = EventManager(self.queue)

    def load_jobs(self):
        for root, dirs, files in os.walk(self.job_directory):
            for name in files:
                if fnmatch.fnmatch(name, 'job.ready'):
                    self.queue.add_job(root)
        
    @property
    def event_handler(self):
        eh = PatternMatchingEventHandler(
            self.patterns, self.ignore_patterns, 
            self.ignore_directories, self.case_sensitive)
        eh.on_created = self.em.on_created
        return eh

    @property
    def observer(self):
        o = Observer()
        o.schedule(
            self.event_handler, self.job_directory, 
            recursive=self.go_recursively)
        return o

    def run(self):
        self.load_jobs()
        self.queue.run_workers(1)
        observer = self.observer
        observer.start()
        try:
            #fallback to stop the observer
            while True:
                log = logging.getLogger('runners')
                log.info(self.queue.status)
                time.sleep(15)
                sys.stdout.flush()
                sys.stderr.flush()
        except KeyboardInterrupt:
            observer.stop()
            observer.join()
            self.queue.stop_workers()



from metaworkspace.workspace import MetacityWorkspace
import os

if "METACITYWS" in os.environ:
    MWSPATH = os.environ["METACITYWS"]
else:
    raise Exception("Trying to import or execute api.py without environment properly set.")

#This should be done better, but it's the only option how to avoid creatign the instance upon every request
#Since this is only OOI encapsulating a couple of files, it's arguably okay
mws = MetacityWorkspace(MWSPATH)




from metaworkspace.runtime.processing import runners
from metaworkspace.runtime.workspace import mws

if __name__ == "__main__":
    print("Starting Jobs...")
    print(f"{mws.path}")
    manager = runners.JobManager(mws.jobs_dir)
    manager.run()



import os
from metaworkspace import filesystem as fs
from metacity.datamodel.project import Project
from passlib.context import CryptContext
from secrets import token_bytes
from base64 import b64encode
import binascii

class UserAuthenticator:
    def __init__(self, workspace_path: str):
        self.user_dir = fs.user_dir(workspace_path)
        self.config = fs.read_json(fs.security_config(workspace_path))
        self.pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")

    @staticmethod
    def generate(workspace_path: str):
        config = fs.security_config(workspace_path)
        fs.write_json(config, {
            'secret': binascii.hexlify(token_bytes(32)).decode("ascii")
        })

    def get_user(self, username):
        file = fs.user_file(self.user_dir, username)
        if fs.file_exists(file):
            return fs.read_json(file)
        return None

    @property
    def secret(self):
        return self.config['secret']

    def get_password_hash(self, password):
        return self.pwd_context.hash(password)

    def verify_password(self, plain_password, hashed_password):
        return self.pwd_context.verify(plain_password, hashed_password)

    def create_user(self, username, fullname, password, email):
        file = fs.user_file(self.user_dir, username)
        if fs.file_exists(file):
            raise Exception("User already exists")

        user = {
            'username': username,
            'full_name': fullname,
            'password': self.get_password_hash(password),
            'email': email
        }
        fs.write_json(file, user)

    def verify_user(self, username, password):
        user = self.get_user(username)
        if not user:
            return None
        if not self.verify_password(password, user['password']):
            return None
        return user

    

class MetacityWorkspace:
    def __init__(self, workspace_path: str):
        self.path = os.path.abspath(workspace_path)
        self.security = UserAuthenticator(self.path)

    @property
    def projects_dir(self):
        return fs.projects_dir(self.path)

    @property
    def jobs_dir(self):
        return fs.jobs_dir(self.path)

    @property
    def logs_dir(self):
        return fs.logs_dir(self.path)

    @property
    def project_names(self):
        return fs.get_projects(self.path)

    @property
    def projects(self):
        for prj_dir in fs.get_projects(self.path):
            yield Project(prj_dir)

    def get_project(self, name: str):
        prj = fs.project_dir(self.path, name)
        fs.create_dir_if_not_exists(prj)
        return Project(prj)

    def delete_project(self, name: str):
        prj = fs.project_dir(self.path, name)
        fs.remove_dirtree(prj)

    def rename_project(self, old_name, new_name):
        old_dir = fs.project_dir(self.path, old_name)
        new_dir = fs.project_dir(self.path, new_name)
        return fs.rename(old_dir, new_dir)   

    @property
    def server_log(self):
        return fs.server_log(self.path) 

    @property
    def jobs_log(self):
        return fs.jobs_log(self.path) 

    def generate_job_dir(self):
        return fs.generate_job_dir(self.path)

    def clear_logs(self):
        fs.archive_logs(self.path)

    def get_current_job_configs(self):
        return fs.current_job_configs(self.path)

    def get_log(self, log_name):
        return fs.log_contents(self.path, log_name)

    def get_logs(self):
        return fs.current_logs(self.path)



import os
from typing import List
from time import sleep
from fastapi import APIRouter, File, Form, UploadFile, Depends, HTTPException, status, Response
from fastapi.responses import JSONResponse
from metaworkspace.runtime.processing.jobs.loaddataset import JobLoadDataset
from metaworkspace.runtime.processing.jobs.buildlayout import JobBuildLayout
from metaworkspace.runtime.workspace import mws
from pydantic import BaseModel
from metaworkspace.runtime.api.auth import User, get_current_active_user

router = APIRouter(prefix="/api",
                   tags=["api"])


class ProjectData(BaseModel):
    name: str


class RenameProjectData(BaseModel):
    old: str
    new: str


class LayerData(BaseModel):
    project: str
    name: str


class RenameLayerData(BaseModel):
    project: str
    old: str
    new: str


class LogData(BaseModel):
    name: str


#### PROJECT

@router.get("/projects", response_class=JSONResponse)
async def list_projects():
    projects = []
    for project in mws.project_names:
        projects.append(project)
    return JSONResponse(projects)


@router.post("/project")
async def add_project(project: ProjectData, current_user: User = Depends(get_current_active_user)):
    for p in mws.project_names:
        if p == project.name:
            raise HTTPException(
                status_code=status.HTTP_409_CONFLICT,
                detail=f"Project {project.name} already exists"
            )

    mws.get_project(project.name)
    return Response(status_code=status.HTTP_201_CREATED) 


@router.post("/project/rename")
async def rename_project(project: RenameProjectData, current_user: User = Depends(get_current_active_user)):
    if mws.rename_project(project.old, project.new):
        return Response(status_code=status.HTTP_200_OK) 
    else:
        raise HTTPException(
                status_code=status.HTTP_409_CONFLICT,
                detail=f"Project {project.new} already exists"
            )


@router.delete("/project")
async def delete_project(project: ProjectData, current_user: User = Depends(get_current_active_user)):
    mws.delete_project(project.name)
    return Response(status_code=status.HTTP_200_OK)


#### LAYER

@router.post("/layers", response_class=JSONResponse)
async def list_layers(project: ProjectData, current_user: User = Depends(get_current_active_user)):
    prj = mws.get_project(project.name)    
    data = []
    for l in prj.ilayers:
        data.append({ 'name': l.name, 'size': l.size })
    return JSONResponse(data) 


@router.post("/layer")
def add_layer(project: str = Form(...), 
              layer: str = Form(...), 
              files: List[UploadFile] = File(...),
              current_user: User = Depends(get_current_active_user)):

    #sleep(3)

    job = JobLoadDataset()
    if layer is None:
        layer = os.path.splitext(files[0].filename)[0]
    job_dir = mws.generate_job_dir()
    job.setup(job_dir, project, layer, files)
    job.submit()

    return Response(status_code=status.HTTP_201_CREATED)


@router.post("/layer/rename")
async def rename_project(layer: RenameLayerData, current_user: User = Depends(get_current_active_user)):
    prj = mws.get_project(layer.project)
    if prj.rename_layer(layer.old, layer.new):
        return Response(status_code=status.HTTP_200_OK) 
    else:
        raise HTTPException(
                status_code=status.HTTP_409_CONFLICT,
                detail=f"Layer {layer.new} already exists"
            )


@router.delete("/layer")
async def delete_project(layer: LayerData, current_user: User = Depends(get_current_active_user)):
    prj = mws.get_project(layer.project)
    prj.delete_layer(layer.name)
    return Response(status_code=status.HTTP_200_OK)


##### LOGS

@router.get("/jobs", response_class=JSONResponse)
async def list_jobs(current_user: User = Depends(get_current_active_user)):
    data = mws.get_current_job_configs()
    return JSONResponse(data) 


@router.get("/logs", response_class=JSONResponse)
async def list_logs(current_user: User = Depends(get_current_active_user)):
    data = mws.get_logs()
    return JSONResponse(data) 


@router.post("/log", response_class=JSONResponse)
async def get_log(log: LogData, current_user: User = Depends(get_current_active_user)):
    data = mws.get_log(log.name)
    if data == None:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Log not found"
        )

    return JSONResponse(data) 

##### ACTIONS

@router.post("/project/build")
def add_layer(project: ProjectData, current_user: User = Depends(get_current_active_user)):

    job = JobBuildLayout()
    job_dir = mws.generate_job_dir()
    job.setup(job_dir, project.name)
    job.submit()
    
    return Response(status_code=status.HTTP_200_OK)




import os
import subprocess
from argparse import ArgumentParser
from getpass import getpass

import metaworkspace.filesystem as fs
from metaworkspace.workspace import MetacityWorkspace, UserAuthenticator


def install(workspace_path: str):
    print(f"Creating workspace directory...")
    fs.recreate_workspace(workspace_path)
    print(f"Generating secrets...")
    UserAuthenticator.generate(workspace_path)
    print(f"Done.")


def regen(workspace_path: str):
    secret = fs.security_config(workspace_path)
    if fs.file_exists(secret):
        os.remove(secret)
    print(f"Generating secrets...")
    UserAuthenticator.generate(workspace_path)
    print(f"Done.")

def createuser(mws: MetacityWorkspace):
    print(f"Generating new user...")
    username =    input("username:              ")
    fullname =    input("full name:             ")
    email =       input("email:                 ")
    while True:
        pass1 = getpass("password (first time): ")
        pass2 = getpass("password (again):      ")
        if pass1 != pass2:
            print("Passwords do not match, try again.")
        else:
            break
    mws.security.create_user(username, fullname, pass1, email)
    print(f"Created user {username}. GLHF! =^.^=")


def run(mws: MetacityWorkspace, ip_adress, port):
    print(f"The server runs on {ip_adress}:{port}")
    print(f"For more information, see logs located in directory {mws.logs_dir}.")
    print(f"Running server...")
    os.environ["METACITYWS"] = mws.path
    mws.clear_logs()

    server_log = open(mws.server_log, 'w+')
    jobs_log = open(mws.jobs_log, 'w+')
    
    proc = subprocess.Popen(["uvicorn", "metaworkspace.runtime.api.server:app", "--workers", "6", "--host", str(ip_adress), "--port", str(port)], 
                            stdout=server_log,
                            stderr=server_log)

    jobs = subprocess.Popen(["python", "-m", "metaworkspace.runtime.processing"], 
                            stdout=jobs_log,
                            stderr=jobs_log)
                 
    return_code = proc.wait()
    return_code = jobs.wait()


usage = ("Sets up Metacity Workspace."
         "Downloads all Metacity modules, sets up local installation of Java,"
         "GMLTools")


if __name__ == "__main__":
    parser = ArgumentParser(description=usage)
    parser.add_argument('--install', help='Run the installation process', action="store_true")
    parser.add_argument('--run', help='Run the application', action="store_true")
    parser.add_argument('--regen', help='Re-generate the secret of the app, it will disable access to all existing users', action="store_true")
    parser.add_argument('--createuser', help='Create new user', action="store_true")
    parser.add_argument('--ip', nargs=1, help='IP adress to run on', default=['127.0.0.1'])
    parser.add_argument('--port', nargs=1, help='Port to run on', default=['5000'])
    parser.add_argument('workspace_dir', type=str, help='Path to newly created Metacity workspace')
    args = parser.parse_args()    
    path = args.workspace_dir

    if args.install:
        install(path)

    if args.regen:
        regen(path)

    try:
        mws = MetacityWorkspace(path)
    except Exception as e:
        print(e)
        print("Could not open the environemnt, please check if the name is right and the secret key is present.")
        quit()

    if args.createuser:
        createuser(mws)

    if args.run:
        run(mws, args.ip[0], args.port[0])





import json
import logging
import os
import shutil
from metaworkspace.runtime.logging import setup_logging
from abc import ABC, abstractmethod
import metaworkspace.filesystem as fs


class Job(ABC):
    TYPE = 'base'
    
    def __init__(self):
        super().__init__()
        self.job_dir = None
        self.job_id = None
        self.log_id = None
        self.status = 'created'
    
    #########################################
    # Job Interface:
    def setup(self, job_dir):
        self.job_dir = job_dir
        self.job_id = fs.id_from_jobdir(self.job_dir)

    def serialize(self):
        return {
            'type': self.TYPE, 
            'status': self.status,
            'job_id': self.job_id,
            'log_id': self.log_id
            }

    @abstractmethod
    def deserialize(self):
        pass

    @abstractmethod
    def run(self):
        pass

    #########################################
    @property
    def job_file(self):
        return os.path.join(self.job_dir, "job.json")
    
    @property
    def job_flag(self):
        return os.path.join(self.job_dir, "job.ready")

    def cleanup(self):
        log = logging.getLogger(self.log_id)
        log.info(f"Job {os.path.basename(self.job_dir)}: {self.serialize()}")
        shutil.rmtree(self.job_dir)

    def submit(self):
        self.submit_checks()
        data = self.serialize()
        with open(self.job_file, 'w') as file:
            json.dump(data, file)
        with open(self.job_flag, 'w') as file:
            pass

    def update_status(self, status):
        self.status = status
        data = self.serialize()
        self.log(self.status)
        with open(self.job_file, 'w') as file:
            json.dump(data, file)

    def log(self, message):
        log = logging.getLogger(self.log_id)
        log.info(f"Job {os.path.basename(self.job_dir)}: {message}")

    def submit_checks(self):
        if not self.job_dir:
            raise Exception("Cannot submit job without assigned directory")

    def setup_log(self, logid):
        self.log_id = logid
        setup_logging(logid)


from typing import Dict, Type
from metaworkspace.runtime.processing.jobs.job import Job
from metaworkspace.runtime.processing.jobs.loaddataset import JobLoadDataset
from metaworkspace.runtime.processing.jobs.buildlayout import JobBuildLayout
import json
import os
import uuid


jobs: Dict[str, Type[Job]] = {
    JobLoadDataset.TYPE: JobLoadDataset,
    JobBuildLayout.TYPE: JobBuildLayout
}


def job_object_hook(data):
    if 'type' not in data:
        raise Exception(f"Job {data} missing job type")
    
    t = data['type']
    if t not in jobs:
        raise Exception(f"Uknown job type: {t}")

    job = jobs[t]()
    job.deserialize(data)
    return job
    
     
def load_job(job_dir):
    jobfile = os.path.join(job_dir, 'job.json')
    with open(jobfile, 'r') as job_file:
        job: Job = json.load(job_file, object_hook=job_object_hook)
    job.job_dir = job_dir
    job.setup_log(str(uuid.uuid1()))
    return job



